use crate::{Statement, Expression, Token, TokenKind, Lexer, Type, Parameter};

pub type Program = Vec<Statement>;
type BindingPower = u8;

#[derive(Debug)]
pub struct Parser<'p> {
    lexer: Lexer<'p>,
    current: Token,
    peek: Token,
}

#[allow(dead_code)]
impl<'p> Parser<'p> {
    pub fn new(lexer: Lexer<'p>) -> Self {
        Self { lexer, current: Token::eof(), peek: Token::eof() }
    }

    pub fn read(&mut self) {
        self.current = self.peek.clone();
        self.peek = if let Some(t) = self.lexer.next() { t } else { Token::eof() };
    }

    fn parse_statement(&mut self) -> Statement {
        match self.current.kind {
            TokenKind::Let => self.parse_let(),
            TokenKind::Fn => self.parse_fn(),
            TokenKind::If => self.parse_if(),
            TokenKind::While => self.parse_while(),
            TokenKind::Return => {
                self.read();

                let expression = self.expression(0);

                self.expect(TokenKind::SemiColon);

                Statement::Return { expression }
            },
            _ => {
                let statement = Statement::Expression { expression: self.expression(0) };
                
                self.expect(TokenKind::SemiColon);
                
                statement
            },
        }
    }

    fn parse_let(&mut self) -> Statement {
        self.read();

        let identifier = self.identifier();
        let r#type = self.r#type();

        self.expect(TokenKind::Equals);

        let statement = Statement::Let { identifier, r#type, initial: self.expression(0) };

        self.expect(TokenKind::SemiColon);

        statement
    }

    fn parse_fn(&mut self) -> Statement {
        self.read();

        let identifier = self.identifier();

        self.expect(TokenKind::LeftParen);

        let parameters = self.parameters();

        self.expect(TokenKind::RightParen);

        let return_type = self.r#type();

        self.expect(TokenKind::LeftBrace);

        let body = self.block(TokenKind::RightBrace);

        self.expect(TokenKind::RightBrace);

        Statement::Function {
            identifier, parameters, return_type, body
        }
    }

    fn parse_if(&mut self) -> Statement {
        self.read();

        let condition = self.expression(0);

        self.expect(TokenKind::LeftBrace);

        let then = self.block(TokenKind::RightBrace);

        self.expect(TokenKind::RightBrace);

        let mut otherwise = Vec::new();

        if self.current.kind == TokenKind::Else {
            self.read();
            self.expect(TokenKind::LeftBrace);

            otherwise = self.block(TokenKind::RightBrace);

            self.expect(TokenKind::RightBrace);
        }

        Statement::If { condition, then, otherwise }
    }

    fn parse_while(&mut self) -> Statement {
        self.read();

        let condition = self.expression(0);

        self.expect(TokenKind::LeftBrace);

        let then = self.block(TokenKind::RightBrace);

        self.expect(TokenKind::RightBrace);

        Statement::While { condition, then }
    }

    fn expect(&mut self, kind: TokenKind) {
        if kind == self.current.kind {
            self.read();
        } else {
            panic!("Expectation failed. {:?}", kind)
        }
    }

    fn expression(&mut self, bp: u8) -> Expression {
        let mut lhs = match self.current.kind.clone() {
            TokenKind::Number(n) => {
                self.read();

                Expression::Number(n)
            },
            TokenKind::True => {
                self.read();

                Expression::Bool(true)
            },
            TokenKind::False => {
                self.read();

                Expression::Bool(false)
            },
            TokenKind::String(s) => {
                self.read();

                Expression::String(s)
            },
            TokenKind::Identifier(i) => {
                self.read();

                Expression::Identifier(i)
            },
            _ if is_prefix(&self.current.kind) => {
                let kind = self.current.kind.clone();

                self.read();

                let (_, rbp) = prefix_binding_power(&kind);

                let rhs = self.expression(rbp);

                prefix(&kind, rhs)
            },
            _ => todo!("{:?}", self.current.kind),
        };

        loop {
            if self.current.kind == TokenKind::Eof {
                break;
            }

            let op = self.current.kind.clone();

            if let Some((lbp, _)) = postfix_binding_power(&op) {
                if lbp < bp {
                    break;
                }

                self.read();

                lhs = postfix(self, lhs, &op);

                continue;
            }

            if let Some((lbp, rbp)) = infix_binding_power(&op) {
                if lbp < bp {
                    break;
                }

                self.read();

                let rhs = self.expression(rbp);

                lhs = infix(lhs, &op, rhs);

                continue;
            }

            break;
        }

        lhs
    }

    fn identifier(&mut self) -> String {
        match self.current.kind.clone() {
            TokenKind::Identifier(i) => {
                self.read();

                i.to_string()
            },
            _ => panic!()
        }
    }

    fn args(&mut self) -> Vec<Expression> {
        let mut args = Vec::new();

        loop {
            if self.current.kind == TokenKind::RightParen {
                break;
            }

            let expression = self.expression(0);

            args.push(expression);

            if self.current.kind == TokenKind::Comma {
                self.read();
            }
        }

        args
    }

    fn parameters(&mut self) -> Vec<Parameter> {
        let mut parameters = Vec::new();

        loop {
            if self.current.kind == TokenKind::RightParen {
                break;
            }

            let identifier = self.identifier();
            let r#type = self.r#type();

            parameters.push(Parameter::new(identifier, r#type));

            if self.current.kind == TokenKind::Comma {
                self.read();
            }
        }

        parameters
    }

    fn block(&mut self, end: TokenKind) -> Vec<Statement> {
        let mut block = Vec::new();

        while self.current.kind != end {
            block.push(self.parse_statement());
        }

        block
    }

    fn r#type(&mut self) -> Option<Type> {
        if self.current.kind != TokenKind::Colon && self.current.kind != TokenKind::DoubleColon {
            None
        } else {
            self.read();

            let r#type = self.identifier();
            
            if ! Type::valid(&r#type) {
                return None
            }

            Some(Type::string(r#type))
        }
    }

    pub fn parse(&mut self) -> Program {
        let mut program = Vec::new();

        while self.current.kind != TokenKind::Eof {
            program.push(self.parse_statement());
        }

        program
    }
}

fn is_prefix(kind: &TokenKind) -> bool {
    [TokenKind::Minus].contains(kind)
}

fn prefix_binding_power(kind: &TokenKind) -> ((), u8) {
    match kind {
        TokenKind::Minus => ((), 99),
        _ => unreachable!()
    }
}

fn prefix(kind: &TokenKind, rhs: Expression) -> Expression {
    Expression::Prefix(kind.clone().into(), Box::new(rhs))
}

fn infix_binding_power(kind: &TokenKind) -> Option<(BindingPower, BindingPower)> {
    Some(match kind {
        TokenKind::Asterisk | TokenKind::Slash | TokenKind::Percent | TokenKind::DoubleAsterisk => (13, 14),
        TokenKind::Plus | TokenKind::Minus => (11, 12),
        TokenKind::GreaterThan | TokenKind::GreaterThanEquals | TokenKind::LessThan | TokenKind::LessThanEquals => (9, 10),
        TokenKind::EqualsEquals | TokenKind::NotEquals => (7, 8),
        TokenKind::Equals => (2, 1),
        _ => return None
    })
}

fn infix(lhs: Expression, kind: &TokenKind, rhs: Expression) -> Expression {
    Expression::Infix(lhs.boxed(), kind.clone().into(), rhs.boxed())
}

fn postfix_binding_power(kind: &TokenKind) -> Option<(BindingPower, ())> {
    Some(match kind {
        TokenKind::LeftParen => (19, ()),
        _ => return None
    })
}

fn postfix(parser: &mut Parser, lhs: Expression, kind: &TokenKind) -> Expression {
    match kind {
        TokenKind::LeftParen => {
            let args = parser.args();

            parser.read();

            Expression::Call(lhs.boxed(), args)
        },
        _ => todo!()
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::Op;

    #[test]
    fn let_statements() {
        assert_eq!(parse("let name = 1;"), vec![
            Statement::Let {
                identifier: String::from("name"),
                r#type: None,
                initial: Expression::Number(1.0),
            },
        ]);

        assert_eq!(parse("let name: number = 1;"), vec![
            Statement::Let {
                identifier: String::from("name"),
                r#type: Some(Type::Number),
                initial: Expression::Number(1.0),
            },
        ]);
    }

    #[test]
    fn fn_statements() {
        assert_eq!(parse("fn name() {}"), vec![
            Statement::Function {
                identifier: String::from("name"),
                parameters: Vec::new(),
                return_type: None,
                body: Vec::new(),
            }
        ]);

        assert_eq!(parse("fn name() :: number {}"), vec![
            Statement::Function {
                identifier: String::from("name"),
                parameters: Vec::new(),
                return_type: Some(Type::Number),
                body: Vec::new(),
            }
        ]);

        assert_eq!(parse("fn name(hello) {}"), vec![
            Statement::Function {
                identifier: String::from("name"),
                parameters: vec![
                    Parameter::new("hello", None),
                ],
                return_type: None,
                body: Vec::new(),
            }
        ]);

        assert_eq!(parse("fn name(hello: string) {}"), vec![
            Statement::Function {
                identifier: String::from("name"),
                parameters: vec![
                    Parameter::new("hello", Some(Type::String)),
                ],
                return_type: None,
                body: Vec::new(),
            }
        ]);

        assert_eq!(parse("fn name(hello: string) :: string {}"), vec![
            Statement::Function {
                identifier: String::from("name"),
                parameters: vec![
                    Parameter::new("hello", Some(Type::String)),
                ],
                return_type: Some(Type::String),
                body: Vec::new(),
            }
        ]);

        assert_eq!(parse(r##"
            fn name(hello: string) :: string {
                let name = "testing";
            }"##
        ), vec![
            Statement::Function {
                identifier: String::from("name"),
                parameters: vec![
                    Parameter::new("hello", Some(Type::String)),
                ],
                return_type: Some(Type::String),
                body: vec![
                    Statement::Let { identifier: String::from("name"), r#type: None, initial: Expression::String("testing".into()) },
                ],
            }
        ]);
    }

    #[test]
    fn returns() {
        assert_eq!(parse("return true;"), vec![
            Statement::Return {
                expression: Expression::Bool(true),
            }
        ]);
    }

    #[test]
    fn if_statements() {
        assert_eq!(parse("if true {}"), vec![
            Statement::If {
                condition: Expression::Bool(true),
                then: vec![],
                otherwise: vec![],
            }
        ]);

        assert_eq!(parse("if true {} else {}"), vec![
            Statement::If {
                condition: Expression::Bool(true),
                then: vec![],
                otherwise: vec![],
            }
        ]);

        assert_eq!(parse("
            if true {
                let age = 1;
            }
        "), vec![
            Statement::If {
                condition: Expression::Bool(true),
                then: vec![
                    Statement::Let { identifier: String::from("age"), r#type: None, initial: Expression::Number(1.0) }
                ],
                otherwise: vec![],
            }
        ]);

        assert_eq!(parse("
            if true {

            } else {
                let age = 1;
            }
        "), vec![
            Statement::If {
                condition: Expression::Bool(true),
                then: vec![],
                otherwise: vec![
                    Statement::Let { identifier: String::from("age"), r#type: None, initial: Expression::Number(1.0) }
                ],
            }
        ]);
    }

    #[test]
    fn while_statements() {
        assert_eq!(parse("while true {}"), vec![
            Statement::While { condition: Expression::Bool(true), then: vec![] }
        ]);

        assert_eq!(parse("while true { 1; }"), vec![
            Statement::While { condition: Expression::Bool(true), then: vec![
                Statement::Expression { expression: Expression::Number(1.0) },
            ] }
        ]);
    }

    #[test]
    fn literals() {
        assert_eq!(parse(r##"1; true; false; "testing";"##), vec![
            Statement::Expression { expression: Expression::Number(1.0) },
            Statement::Expression { expression: Expression::Bool(true) },
            Statement::Expression { expression: Expression::Bool(false) },
            Statement::Expression { expression: Expression::String(String::from("testing")) },
        ]);
    }

    #[test]
    fn prefixes() {
        assert_eq!(parse("-1;"), vec![
            Statement::Expression { expression: Expression::Prefix(Op::Subtract, Box::new(Expression::Number(1.0))) },
        ]);
    }

    #[test]
    fn infixes() {
        assert_eq!(parse("1 + 1;"), vec![
            Statement::Expression { expression: Expression::Infix(Expression::Number(1.0).boxed(), Op::Add, Expression::Number(1.0).boxed()) }
        ]);

        assert_eq!(parse("1 - 1;"), vec![
            Statement::Expression { expression: Expression::Infix(Expression::Number(1.0).boxed(), Op::Subtract, Expression::Number(1.0).boxed()) }
        ]);

        assert_eq!(parse("1 * 1;"), vec![
            Statement::Expression { expression: Expression::Infix(Expression::Number(1.0).boxed(), Op::Multiply, Expression::Number(1.0).boxed()) }
        ]);

        assert_eq!(parse("1 / 1;"), vec![
            Statement::Expression { expression: Expression::Infix(Expression::Number(1.0).boxed(), Op::Divide, Expression::Number(1.0).boxed()) }
        ]);

        assert_eq!(parse("1 + 2 * 3;"), vec![
            Statement::Expression {
                expression: Expression::Infix(
                    Expression::Number(1.0).boxed(), Op::Add, Expression::Infix(
                        Expression::Number(2.0).boxed(),
                        Op::Multiply,
                        Expression::Number(3.0).boxed()
                    ).boxed()
                )
            }
        ]);

        assert_eq!(parse("1 + 2 * 3 / 4;"), vec![
            Statement::Expression {
                expression: Expression::Infix(
                    Expression::Number(1.0).boxed(), Op::Add, Expression::Infix(
                        Expression::Infix(
                            Expression::Number(2.0).boxed(),
                            Op::Multiply,
                            Expression::Number(3.0).boxed()
                        ).boxed(),
                        Op::Divide,
                        Expression::Number(4.0).boxed()
                    ).boxed()
                )
            }
        ]);

        assert_eq!(parse("1 > 1;"), vec![
            Statement::Expression { expression: Expression::Infix(Expression::Number(1.0).boxed(), Op::GreaterThan, Expression::Number(1.0).boxed()) }
        ]);

        assert_eq!(parse("1 < 1;"), vec![
            Statement::Expression { expression: Expression::Infix(Expression::Number(1.0).boxed(), Op::LessThan, Expression::Number(1.0).boxed()) }
        ]);

        assert_eq!(parse("1 >= 1;"), vec![
            Statement::Expression { expression: Expression::Infix(Expression::Number(1.0).boxed(), Op::GreaterThanEquals, Expression::Number(1.0).boxed()) }
        ]);

        assert_eq!(parse("1 <= 1;"), vec![
            Statement::Expression { expression: Expression::Infix(Expression::Number(1.0).boxed(), Op::LessThanEquals, Expression::Number(1.0).boxed()) }
        ]);

        assert_eq!(parse("1 == 1;"), vec![
            Statement::Expression { expression: Expression::Infix(Expression::Number(1.0).boxed(), Op::Equals, Expression::Number(1.0).boxed()) }
        ]);

        assert_eq!(parse("1 != 1;"), vec![
            Statement::Expression { expression: Expression::Infix(Expression::Number(1.0).boxed(), Op::NotEquals, Expression::Number(1.0).boxed()) }
        ]);
        
        assert_eq!(parse("foo = 2;"), vec![
            Statement::Expression {
                expression: Expression::Infix(
                    Expression::Identifier("foo".to_owned()).boxed(),
                    Op::Assign,
                    Expression::Number(2.0).boxed()
                )
            }
        ]);
    }

    #[test]
    fn postfixes() {
        assert_eq!(parse("foo();"), vec![
            Statement::Expression {
                expression: Expression::Call(
                    Expression::Identifier("foo".to_owned()).boxed(),
                    vec![]
                )
            }
        ]);

        assert_eq!(parse("foo() + foo();"), vec![
            Statement::Expression {
                expression: Expression::Infix(
                    Expression::Call(Expression::Identifier("foo".to_owned()).boxed(), Vec::new()).boxed(),
                    Op::Add,
                    Expression::Call(Expression::Identifier("foo".to_owned()).boxed(), Vec::new()).boxed(),
                )
            }
        ]);

        assert_eq!(parse("foo(1);"), vec![
            Statement::Expression {
                expression: Expression::Call(
                    Expression::Identifier("foo".to_owned()).boxed(),
                    vec![
                        Expression::Number(1.0),
                    ]
                )
            }
        ]);

        assert_eq!(parse("foo(1, 2, 3);"), vec![
            Statement::Expression {
                expression: Expression::Call(
                    Expression::Identifier("foo".to_owned()).boxed(),
                    vec![
                        Expression::Number(1.0),
                        Expression::Number(2.0),
                        Expression::Number(3.0),
                    ]
                )
            }
        ]);

        assert_eq!(parse("foo(1, 2, 3,);"), vec![
            Statement::Expression {
                expression: Expression::Call(
                    Expression::Identifier("foo".to_owned()).boxed(),
                    vec![
                        Expression::Number(1.0),
                        Expression::Number(2.0),
                        Expression::Number(3.0),
                    ]
                )
            }
        ]);
    }

    fn parse(source: &str) -> Program {
        let lexer = Lexer::new(source);
        let mut parser = Parser::new(lexer);

        parser.read();
        parser.read();

        parser.parse()
    }
}